@@ -146,6 +146,9 @@ struct pbap_session {
 
 struct pbap_object {
 	GString *buffer;
+	GByteArray *aparams;
+	gboolean firstpacket;
+	gboolean lastpart;
 	struct pbap_session *session;
 	void *request;
 };
@@ -224,12 +227,38 @@ static void cache_clear(struct cache *cache)
 	cache->entries = NULL;
 }
 
+static GByteArray *append_aparam_header(GByteArray *buf, uint8_t tag,
+							const void *val)
+{
+	/* largest aparam is for phonebooksize (4 bytes) */
+	uint8_t aparam[sizeof(struct aparam_header) + PHONEBOOKSIZE_LEN];
+	struct aparam_header *hdr = (struct aparam_header *) aparam;
+
+	switch (tag) {
+	case PHONEBOOKSIZE_TAG:
+		hdr->tag = PHONEBOOKSIZE_TAG;
+		hdr->len = PHONEBOOKSIZE_LEN;
+		memcpy(hdr->val, val, PHONEBOOKSIZE_LEN);
+
+		return g_byte_array_append(buf,	aparam,
+			sizeof(struct aparam_header) + PHONEBOOKSIZE_LEN);
+	case NEWMISSEDCALLS_TAG:
+		hdr->tag = NEWMISSEDCALLS_TAG;
+		hdr->len = NEWMISSEDCALLS_LEN;
+		memcpy(hdr->val, val, NEWMISSEDCALLS_LEN);
+
+		return g_byte_array_append(buf,	aparam,
+			sizeof(struct aparam_header) + NEWMISSEDCALLS_LEN);
+	default:
+		return buf;
+	}
+}
+
 static void phonebook_size_result(const char *buffer, size_t bufsize,
-				int vcards, int missed, void *user_data)
+					int vcards, int missed,
+					gboolean lastpart, void *user_data)
 {
 	struct pbap_session *pbap = user_data;
-	char aparam[4];
-	struct aparam_header *hdr = (struct aparam_header *) aparam;
 	uint16_t phonebooksize;
 
 	if (pbap->obj->request) {
@@ -244,28 +273,35 @@ static void phonebook_size_result(const char *buffer, size_t bufsize,
 
 	phonebooksize = htons(vcards);
 
-	hdr->tag = PHONEBOOKSIZE_TAG;
-	hdr->len = PHONEBOOKSIZE_LEN;
-	memcpy(hdr->val, &phonebooksize, sizeof(phonebooksize));
+	pbap->obj->aparams = g_byte_array_new();
+	pbap->obj->aparams = append_aparam_header(pbap->obj->aparams,
+					PHONEBOOKSIZE_TAG, &phonebooksize);
 
-	pbap->obj->buffer = g_string_new_len(aparam, sizeof(aparam));
+	if (missed > 0)	{
+		DBG("missed %d", missed);
+
+		pbap->obj->aparams = append_aparam_header(pbap->obj->aparams,
+						NEWMISSEDCALLS_TAG, &missed);
+	}
 
 	obex_object_set_io_flags(pbap->obj, G_IO_IN, 0);
 }
 
 static void query_result(const char *buffer, size_t bufsize, int vcards,
-					int missed, void *user_data)
+				int missed, gboolean lastpart, void *user_data)
 {
 	struct pbap_session *pbap = user_data;
 
 	DBG("");
 
-	if (pbap->obj->request) {
+	if (pbap->obj->request && lastpart) {
 		phonebook_req_finalize(pbap->obj->request);
 		pbap->obj->request = NULL;
 	}
 
-	if (vcards <= 0) {
+	pbap->obj->lastpart = lastpart;
+
+	if (vcards < 0) {
 		obex_object_set_io_flags(pbap->obj, G_IO_ERR, -ENOENT);
 		return;
 	}
@@ -276,6 +312,16 @@ static void query_result(const char *buffer, size_t bufsize, int vcards,
 		pbap->obj->buffer = g_string_append_len(pbap->obj->buffer,
 							buffer,	bufsize);
 
+	if (missed > 0)	{
+		DBG("missed %d", missed);
+
+		pbap->obj->firstpacket = TRUE;
+
+		pbap->obj->aparams = g_byte_array_new();
+		pbap->obj->aparams = append_aparam_header(pbap->obj->aparams,
+						NEWMISSEDCALLS_TAG, &missed);
+	}
+
 	obex_object_set_io_flags(pbap->obj, G_IO_IN, 0);
 }
 
@@ -400,15 +446,11 @@ static int generate_response(void *user_data)
 
 	if (max == 0) {
 		/* Ignore all other parameter and return PhoneBookSize */
-		char aparam[4];
-		struct aparam_header *hdr = (struct aparam_header *) aparam;
 		uint16_t size = htons(g_slist_length(pbap->cache.entries));
 
-		hdr->tag = PHONEBOOKSIZE_TAG;
-		hdr->len = PHONEBOOKSIZE_LEN;
-		memcpy(hdr->val, &size, sizeof(size));
-
-		pbap->obj->buffer = g_string_new_len(aparam, sizeof(aparam));
+		pbap->obj->aparams = g_byte_array_new();
+		pbap->obj->aparams = append_aparam_header(pbap->obj->aparams,
+						PHONEBOOKSIZE_TAG, &size);
 
 		return 0;
 	}
@@ -420,8 +462,6 @@ static int generate_response(void *user_data)
 	sorted = sort_entries(pbap->cache.entries, pbap->params->order,
 				pbap->params->searchattrib,
 				(const char *) pbap->params->searchval);
-	if (sorted == NULL)
-		return -ENOENT;
 
 	/* Computing offset considering first entry of the phonebook */
 	l = g_slist_nth(sorted, pbap->params->liststartoffset);
@@ -444,18 +484,15 @@ static int generate_response(void *user_data)
 static void cache_ready_notify(void *user_data)
 {
 	struct pbap_session *pbap = user_data;
-	int err;
 
 	DBG("");
 
+	phonebook_req_finalize(pbap->obj->request);
+	pbap->obj->request = NULL;
+
 	pbap->cache.valid = TRUE;
 
-	err = generate_response(pbap);
-	if (err < 0) {
-		obex_object_set_io_flags(pbap->obj, G_IO_ERR, err);
-		return;
-	}
-
+	generate_response(pbap);
 	obex_object_set_io_flags(pbap->obj, G_IO_IN, 0);
 }
 
@@ -476,9 +513,7 @@ static void cache_entry_done(void *user_data)
 		return;
 	}
 
-	/* Unref previous request, associated data will be freed. */
 	phonebook_req_finalize(pbap->obj->request);
-	/* Get new pointer to pending call. */
 	pbap->obj->request = phonebook_get_entry(pbap->folder, id,
 				pbap->params, query_result, pbap, &ret);
 	if (ret < 0)
@@ -585,7 +620,7 @@ static void *pbap_connect(struct obex_session *os, int *err)
 }
 
 static int pbap_get(struct obex_session *os, obex_object_t *obj,
-					gboolean *stream, void *user_data)
+							void *user_data)
 {
 	struct pbap_session *pbap = user_data;
 	const char *type = obex_get_type(os);
@@ -618,8 +653,11 @@ static int pbap_get(struct obex_session *os, obex_object_t *obj,
 
 	if (strcmp(type, PHONEBOOK_TYPE) == 0) {
 		/* Always contains the absolute path */
-		path = g_strdup(name);
-		*stream = (params->maxlistcount == 0 ? FALSE : TRUE);
+		if (g_path_is_absolute(name))
+			path = g_strdup(name);
+		else
+			path = g_build_filename("/", name, NULL);
+
 	} else if (strcmp(type, VCARDLISTING_TYPE) == 0) {
 		/* Always relative */
 		if (!name || strlen(name) == 0)
@@ -629,11 +667,9 @@ static int pbap_get(struct obex_session *os, obex_object_t *obj,
 			/* Current folder + relative path */
 			path = g_build_filename(pbap->folder, name, NULL);
 
-		*stream = (params->maxlistcount == 0 ? FALSE : TRUE);
 	} else if (strcmp(type, VCARDENTRY_TYPE) == 0) {
 		/* File name only */
 		path = g_strdup(name);
-		*stream = TRUE;
 	} else
 		return -EBADR;
 
@@ -763,6 +799,11 @@ static void *vobject_pull_open(const char *name, int oflag, mode_t mode,
 
 	request = phonebook_pull(name, pbap->params, cb, pbap, &ret);
 
+	if (ret < 0)
+		goto fail;
+
+	/* reading first part of results from backend */
+	ret = phonebook_pull_read(request);
 	if (ret < 0)
 		goto fail;
 
@@ -790,10 +831,11 @@ static int vobject_close(void *object)
 	if (obj->buffer)
 		g_string_free(obj->buffer, TRUE);
 
-	if (obj->request) {
+	if (obj->aparams)
+		g_byte_array_free(obj->aparams, TRUE);
+
+	if (obj->request)
 		phonebook_req_finalize(obj->request);
-		obj->request = NULL;
-	}
 
 	g_free(obj);
 
@@ -902,51 +944,104 @@ fail:
 	return NULL;
 }
 
-static ssize_t vobject_pull_read(void *object, void *buf, size_t count,
+static ssize_t array_read(GByteArray *array, void *buf, size_t count)
+{
+	ssize_t len;
+
+	if (array->len == 0)
+		return 0;
+
+	len = MIN(array->len, count);
+	memcpy(buf, array->data, len);
+	array = g_byte_array_remove_range(array, 0, len);
+
+	return len;
+}
+
+static ssize_t vobject_pull_get_next_header(void *object, void *buf, size_t mtu,
 								uint8_t *hi)
 {
 	struct pbap_object *obj = object;
 	struct pbap_session *pbap = obj->session;
 
+	if (!obj->buffer && !obj->aparams)
+		return -EAGAIN;
+
+	*hi = OBEX_HDR_APPARAM;
+
+	if (pbap->params->maxlistcount == 0 || obj->firstpacket) {
+		obj->firstpacket = FALSE;
+
+		return array_read(obj->aparams, buf, mtu);
+	}
+
+	return 0;
+}
+
+static ssize_t vobject_pull_read(void *object, void *buf, size_t count)
+{
+	struct pbap_object *obj = object;
+	struct pbap_session *pbap = obj->session;
+	int len, ret;
+
 	DBG("buffer %p maxlistcount %d", obj->buffer,
 						pbap->params->maxlistcount);
 
-	if (!obj->buffer)
+	if (!obj->buffer && !obj->aparams)
 		return -EAGAIN;
 
-	/* PhoneBookSize */
 	if (pbap->params->maxlistcount == 0)
-		*hi = OBEX_HDR_APPARAM;
-	else
-		/* Stream data */
-		*hi = OBEX_HDR_BODY;
+		return -ENOSTR;
 
-	return string_read(obj->buffer, buf, count);
+	len = string_read(obj->buffer, buf, count);
+	if (len == 0 && !obj->lastpart) {
+		/* in case when buffer is empty and we know that more
+		 * data is still available in backend, requesting new
+		 * data part via phonebook_pull_read and returning
+		 * -EAGAIN to suspend request for now */
+		ret = phonebook_pull_read(obj->request);
+		if (ret)
+			return -EPERM;
+
+		return -EAGAIN;
+	}
+
+	return len;
 }
 
-static ssize_t vobject_list_read(void *object, void *buf, size_t count,
+static ssize_t vobject_list_get_next_header(void *object, void *buf, size_t mtu,
 								uint8_t *hi)
 {
 	struct pbap_object *obj = object;
 	struct pbap_session *pbap = obj->session;
 
+	/* Backend still busy reading contacts */
+	if (!pbap->cache.valid)
+		return -EAGAIN;
+
+	*hi = OBEX_HDR_APPARAM;
+
+	if (pbap->params->maxlistcount == 0)
+		return array_read(obj->aparams, buf, mtu);
+
+	return 0;
+}
+
+static ssize_t vobject_list_read(void *object, void *buf, size_t count)
+{
+	struct pbap_object *obj = object;
+	struct pbap_session *pbap = obj->session;
+
 	DBG("valid %d maxlistcount %d", pbap->cache.valid,
 						pbap->params->maxlistcount);
 
-	/* Backend still busy reading contacts */
-	if (!pbap->cache.valid)
-		return -EAGAIN;
-
 	if (pbap->params->maxlistcount == 0)
-		*hi = OBEX_HDR_APPARAM;
-	else
-		*hi = OBEX_HDR_BODY;
+		return -ENOSTR;
 
 	return string_read(obj->buffer, buf, count);
 }
 
-static ssize_t vobject_vcard_read(void *object, void *buf, size_t count,
-								uint8_t *hi)
+static ssize_t vobject_vcard_read(void *object, void *buf, size_t count)
 {
 	struct pbap_object *obj = object;
 
@@ -955,7 +1050,6 @@ static ssize_t vobject_vcard_read(void *object, void *buf, size_t count,
 	if (!obj->buffer)
 		return -EAGAIN;
 
-	*hi = OBEX_HDR_BODY;
 	return string_read(obj->buffer, buf, count);
 }
 
@@ -966,6 +1060,7 @@ static struct obex_mime_type_driver mime_pull = {
 	.open = vobject_pull_open,
 	.close = vobject_close,
 	.read = vobject_pull_read,
+	.get_next_header = vobject_pull_get_next_header,
 };
 
 static struct obex_mime_type_driver mime_list = {
@@ -975,6 +1070,7 @@ static struct obex_mime_type_driver mime_list = {
 	.open = vobject_list_open,
 	.close = vobject_close,
 	.read = vobject_list_read,
+	.get_next_header = vobject_list_get_next_header,
 };
 
 static struct obex_mime_type_driver mime_vcard = {
@@ -992,7 +1088,7 @@ static int pbap_init(void)
 
 	err = phonebook_init();
 	if (err < 0)
-		goto fail_pb_init;
+		return err;
 
 	err = obex_mime_type_driver_register(&mime_pull);
 	if (err < 0)
@@ -1020,7 +1116,7 @@ fail_mime_list:
 	obex_mime_type_driver_unregister(&mime_pull);
 fail_mime_pull:
 	phonebook_exit();
-fail_pb_init:
+
 	return err;
 }
 
