@@ -45,18 +45,20 @@
 #define QUERY_NAME "(contains \"given_name\" \"%s\")"
 #define QUERY_PHONE "(contains \"phone\" \"%s\")"
 
-
 struct query_context {
-	gboolean completed;
 	const struct apparam_field *params;
 	phonebook_cb contacts_cb;
 	phonebook_entry_cb entry_cb;
 	phonebook_cache_ready_cb ready_cb;
+	EBookQuery *query;
+	unsigned int count;
+	GString *buf;
+	char *id;
+	unsigned queued_calls;
 	void *user_data;
+	GSList *ebooks;
 };
 
-static EBook *ebook = NULL;
-
 static char *attribute_mask[] = {
 /* 0 */		"VERSION",
 		"FN",
@@ -91,6 +93,26 @@ static char *attribute_mask[] = {
 
 };
 
+static void close_ebooks(GSList *ebooks)
+{
+	g_slist_free_full(ebooks, g_object_unref);
+}
+
+static void free_query_context(struct query_context *data)
+{
+	g_free(data->id);
+
+	if (data->buf != NULL)
+		g_string_free(data->buf, TRUE);
+
+	if (data->query != NULL)
+		e_book_query_unref(data->query);
+
+	close_ebooks(data->ebooks);
+
+	g_free(data);
+}
+
 static char *evcard_to_string(EVCard *evcard, unsigned int format,
 							uint64_t filter)
 {
@@ -99,7 +121,8 @@ static char *evcard_to_string(EVCard *evcard, unsigned int format,
 	char *vcard;
 
 	if (!filter)
-		return e_vcard_to_string(evcard, format);
+		return e_vcard_to_string(evcard, EVC_FORMAT_VCARD_30);
+		/* XXX There is no support for VCARD 2.1 at this time */
 
 	/*
 	 * Mandatory attributes for vCard 2.1 are VERSION ,N and TEL.
@@ -136,24 +159,20 @@ static char *evcard_to_string(EVCard *evcard, unsigned int format,
 	return vcard;
 }
 
-static void ebookpull_cb(EBook *book, EBookStatus estatus, GList *contacts,
+static void ebookpull_cb(EBook *book, const GError *gerr, GList *contacts,
 							void *user_data)
 {
 	struct query_context *data = user_data;
-	GString *string = g_string_new("");
-	unsigned int count = 0, maxcount;
 	GList *l;
+	unsigned int count = 0, maxcount;
 
-	if (estatus == E_BOOK_ERROR_CANCELLED) {
-		error("E-Book operation was cancelled: status %d", estatus);
-		goto fail;
-	}
-
-	if (estatus != E_BOOK_ERROR_OK) {
-		error("E-Book query failed: status %d", estatus);
+	if (gerr != NULL) {
+		error("E-Book query failed: %s", gerr->message);
 		goto done;
 	}
 
+	DBG("");
+
 	/*
 	 * When MaxListCount is zero, PCE wants to know the number of used
 	 * indexes in the phonebook of interest. All other parameters that
@@ -161,15 +180,14 @@ static void ebookpull_cb(EBook *book, EBookStatus estatus, GList *contacts,
 	 */
 	maxcount = data->params->maxlistcount;
 	if (maxcount == 0) {
-		count = g_list_length(contacts);
+		count += g_list_length(contacts);
 		goto done;
 	}
 
 	l = g_list_nth(contacts, data->params->liststartoffset);
 
-	/* FIXME: Missing 0.vcf */
-
-	for (; l && count < maxcount; l = g_list_next(l), count++) {
+	for (; l && count + data->count < maxcount; l = g_list_next(l),
+								count++) {
 		EContact *contact = E_CONTACT(l->data);
 		EVCard *evcard = E_VCARD(contact);
 		char *vcard;
@@ -177,45 +195,44 @@ static void ebookpull_cb(EBook *book, EBookStatus estatus, GList *contacts,
 		vcard = evcard_to_string(evcard, data->params->format,
 						data->params->filter);
 
-		string = g_string_append(string, vcard);
-		string = g_string_append(string, "\r\n");
+		data->buf = g_string_append(data->buf, vcard);
+		data->buf = g_string_append(data->buf, "\r\n");
 		g_free(vcard);
 	}
-	DBG("collected %d vcards", count);
 
+	data->count += count;
 
 done:
-	data->completed = TRUE;
-	data->contacts_cb(string->str, string->len, count, 0, TRUE,
-							data->user_data);
+	g_list_free_full(contacts, g_object_unref);
 
-fail:
-	g_string_free(string, TRUE);
+	DBG("collected %d vcards", count);
 
-	if (data->completed)
-		g_free(data);
+	data->queued_calls--;
+	if (data->queued_calls == 0) {
+		GString *buf = data->buf;
+		data->buf = NULL;
+
+		data->contacts_cb(buf->str, buf->len, data->count,
+						0, TRUE, data->user_data);
+
+		g_string_free(buf, TRUE);
+	}
 }
 
-static void ebook_entry_cb(EBook *book, EBookStatus estatus,
-			EContact *contact, void *user_data)
+static void ebook_entry_cb(EBook *book, const GError *gerr,
+				EContact *contact, void *user_data)
 {
 	struct query_context *data = user_data;
 	EVCard *evcard;
 	char *vcard;
 	size_t len;
 
-	if (estatus == E_BOOK_ERROR_CANCELLED) {
-		error("E-Book operation was cancelled: status %d", estatus);
-		goto fail;
+	if (gerr != NULL) {
+		error("E-Book query failed: %s", gerr->message);
+		goto done;
 	}
 
-	data->completed = TRUE;
-
-	if (estatus != E_BOOK_ERROR_OK) {
-		error("E-Book query failed: status %d", estatus);
-		data->contacts_cb(NULL, 0, 1, 0, TRUE, data->user_data);
-		goto fail;
-	}
+	DBG("");
 
 	evcard = E_VCARD(contact);
 
@@ -224,13 +241,21 @@ static void ebook_entry_cb(EBook *book, EBookStatus estatus,
 
 	len = vcard ? strlen(vcard) : 0;
 
+	data->count++;
 	data->contacts_cb(vcard, len, 1, 0, TRUE, data->user_data);
 
 	g_free(vcard);
+	g_object_unref(contact);
 
-fail:
-	if (data->completed)
-		g_free(data);
+done:
+	data->queued_calls--;
+	if (data->queued_calls == 0) {
+		if (data->count == 0)
+			data->contacts_cb(NULL, 0, 1, 0, TRUE,
+						data->user_data);
+
+		free_query_context(data);
+	}
 }
 
 static char *evcard_name_attribute_to_string(EVCard *evcard)
@@ -263,24 +288,19 @@ static char *evcard_name_attribute_to_string(EVCard *evcard)
 	return g_string_free(name, FALSE);
 }
 
-static void cache_cb(EBook *book, EBookStatus estatus, GList *contacts,
+static void cache_cb(EBook *book, const GError *gerr, GList *contacts,
 							void *user_data)
 {
 	struct query_context *data = user_data;
 	GList *l;
 
-	if (estatus == E_BOOK_ERROR_CANCELLED) {
-		error("E-Book operation was cancelled: status %d", estatus);
-		goto fail;
-	}
-
-	data->completed = TRUE;
-
-	if (estatus != E_BOOK_ERROR_OK) {
-		error("E-Book query failed: status %d", estatus);
+	if (gerr != NULL) {
+		error("E-Book operation failed: %s", gerr->message);
 		goto done;
 	}
 
+	DBG("");
+
 	for (l = contacts; l; l = g_list_next(l)) {
 		EContact *contact = E_CONTACT(l->data);
 		EVCard *evcard = E_VCARD(contact);
@@ -295,61 +315,116 @@ static void cache_cb(EBook *book, EBookStatus estatus, GList *contacts,
 		if (!attrib)
 			continue;
 
-		uid =  e_vcard_attribute_get_value(attrib);
+		uid = e_vcard_attribute_get_value(attrib);
 		if (!uid)
 			continue;
 
 		attrib = e_vcard_get_attribute(evcard, EVC_TEL);
-		if (!attrib)
-			continue;
-
-		tel =  e_vcard_attribute_get_value(attrib);
+		if (attrib)
+			tel = e_vcard_attribute_get_value(attrib);
+		else
+			tel = g_strdup("");
 
 		data->entry_cb(uid, PHONEBOOK_INVALID_HANDLE, name, NULL,
 							tel, data->user_data);
+
 		g_free(name);
 		g_free(uid);
 		g_free(tel);
 	}
-done:
-	data->ready_cb(data->user_data);
 
-fail:
-	if (data->completed)
-		g_free(data);
+done:
+	g_list_free_full(contacts, g_object_unref);
+
+	data->queued_calls--;
+	if (data->queued_calls == 0)
+		data->ready_cb(data->user_data);
+}
+
+static GSList *traverse_sources(GSList *ebooks, GSList *sources,
+							char *default_src) {
+	GError *gerr;
+
+	while (sources != NULL) {
+		EBook *ebook = e_book_new(E_SOURCE(sources->data), &gerr);
+		if (ebook == NULL) {
+			error("Can't create user's address book: %s",
+								gerr->message);
+			sources = sources->next;
+
+			g_error_free(gerr);
+			continue;
+		}
+
+		if (g_strcmp0(default_src, e_source_get_uri(
+					E_SOURCE(sources->data))) == 0) {
+			sources = sources->next;
+
+			continue;
+		}
+
+		if (e_book_open(ebook, FALSE, &gerr) == FALSE) {
+			error("Can't open e-book address book: %s",
+							gerr->message);
+			sources = sources->next;
+
+			g_object_unref(ebook);
+			g_error_free(gerr);
+			continue;
+		}
+
+		if (default_src == NULL)
+			default_src = e_source_get_uri(E_SOURCE(sources->data));
+
+		DBG("%s address book opened",
+					e_source_peek_name(sources->data));
+
+		ebooks = g_slist_append(ebooks, ebook);
+
+		sources = sources->next;
+	}
+
+	return ebooks;
 }
 
 int phonebook_init(void)
 {
-	GError *gerr = NULL;
-
-	if (ebook)
-		return 0;
-
 	g_type_init();
 
-	ebook = e_book_new_default_addressbook(&gerr);
-	if (!ebook) {
-		error("Can't create user's default address book: %s",
-				gerr->message);
-		g_error_free(gerr);
-		return -EIO;
-	}
-
-	if (!e_book_open(ebook, FALSE, &gerr)) {
-		error("Can't open e-book address book: %s", gerr->message);
-		g_error_free(gerr);
-		return -EIO;
-	}
-
 	return 0;
 }
 
+static GSList *open_ebooks(void)
+{
+	GError *gerr;
+	ESourceList *src_list;
+	GSList *list;
+	gchar *default_src = NULL;
+	GSList *ebooks = NULL;
+
+	if (e_book_get_addressbooks(&src_list, &gerr) == FALSE) {
+		error("Can't list user's address books: %s", gerr->message);
+		g_error_free(gerr);
+
+		return NULL;
+	}
+
+	list = e_source_list_peek_groups(src_list);
+	while (list) {
+		ESourceGroup *group = E_SOURCE_GROUP(list->data);
+
+		GSList *sources = e_source_group_peek_sources(group);
+
+		ebooks = traverse_sources(ebooks, sources, default_src);
+
+		list = list->next;
+	}
+
+	return ebooks;
+}
+
 void phonebook_exit(void)
 {
-	if (ebook)
-		g_object_unref(ebook);
-	ebook = NULL;
 }
 
 char *phonebook_set_folder(const char *current_folder,
@@ -439,14 +514,19 @@ done:
 void phonebook_req_finalize(void *request)
 {
 	struct query_context *data = request;
+	GSList *ebook = data->ebooks;
 
-	if (!data)
-		return;
+	DBG("");
 
-	if (!data->completed) {
-		data->completed = TRUE;
-		e_book_cancel_async_op(ebook, NULL);
+	while (ebook != NULL) {
+		if (e_book_cancel(ebook->data, NULL) == TRUE)
+			data->queued_calls--;
+
+		ebook = ebook->next;
 	}
+
+	if (data != NULL && data->queued_calls == 0)
+		free_query_context(data);
 }
 
 void *phonebook_pull(const char *name, const struct apparam_field *params,
@@ -454,13 +534,23 @@ void *phonebook_pull(const char *name, const struct apparam_field *params,
 {
 	struct query_context *data;
 
+	if (g_strcmp0("/telecom/pb.vcf", name) != 0) {
+		if (err)
+			*err = -ENOENT;
+
+		return NULL;
+	}
+
 	data = g_new0(struct query_context, 1);
 	data->contacts_cb = cb;
 	data->params = params;
 	data->user_data = user_data;
+	data->buf = g_string_new("");
+	data->query = e_book_query_any_field_contains("");
+	data->ebooks = open_ebooks();
 
 	if (err)
-		*err = 0;
+		*err = data->ebooks == NULL ? -EIO : 0;
 
 	return data;
 }
@@ -468,15 +558,26 @@ void *phonebook_pull(const char *name, const struct apparam_field *params,
 int phonebook_pull_read(void *request)
 {
 	struct query_context *data = request;
-	EBookQuery *query;
+	gboolean ret;
+	GSList *ebook;
 
 	if (!data)
 		return -ENOENT;
 
-	query = e_book_query_any_field_contains("");
-	e_book_async_get_contacts(ebook, query, ebookpull_cb, data);
+	ebook = data->ebooks;
+	while (ebook != NULL) {
+		if (e_book_is_opened(ebook->data) == TRUE) {
+			ret = e_book_get_contacts_async(ebook->data,
+					data->query, ebookpull_cb, data);
+			if (ret == TRUE)
+				data->queued_calls++;
+		}
 
-	e_book_query_unref(query);
+		ebook = ebook->next;
+	}
+
+	if (data->queued_calls == 0)
+		return -ENOENT;
 
 	return 0;
 }
@@ -485,22 +586,31 @@ void *phonebook_get_entry(const char *folder, const char *id,
 				const struct apparam_field *params,
 				phonebook_cb cb, void *user_data, int *err)
 {
+	gboolean ret;
 	struct query_context *data;
+	GSList *ebook;
 
 	data = g_new0(struct query_context, 1);
 	data->contacts_cb = cb;
 	data->params = params;
 	data->user_data = user_data;
+	data->id = g_strdup(id);
+	data->ebooks = open_ebooks();
 
-	if (e_book_async_get_contact(ebook, id, ebook_entry_cb, data)) {
-		g_free(data);
-		if (err)
-			*err = -ENOENT;
-		return NULL;
+	ebook = data->ebooks;
+	while (ebook != NULL) {
+		if (e_book_is_opened(ebook->data) == TRUE) {
+			ret = e_book_get_contact_async(ebook->data, data->id,
+							ebook_entry_cb, data);
+			if (ret == TRUE)
+				data->queued_calls++;
+		}
+
+		ebook = ebook->next;
 	}
 
 	if (err)
-		*err = 0;
+		*err = (data->queued_calls == 0 ? -ENOENT : 0);
 
 	return data;
 }
@@ -511,31 +621,79 @@ void *phonebook_create_cache(const char *name, phonebook_entry_cb entry_cb,
 	struct query_context *data;
 	EBookQuery *query;
 	gboolean ret;
+	GSList *ebook;
+	EContact *me;
+	EVCard *evcard;
+	GError *gerr;
+	EBook *eb;
+	EVCardAttribute *attrib;
+	char *uid, *tel, *cname;
 
 	if (g_strcmp0("/telecom/pb", name) != 0) {
 		if (err)
 			*err = -ENOENT;
+
 		return NULL;
 	}
 
+	DBG("");
+
 	query = e_book_query_any_field_contains("");
 
 	data = g_new0(struct query_context, 1);
 	data->entry_cb = entry_cb;
 	data->ready_cb = ready_cb;
 	data->user_data = user_data;
+	data->query = query;
+	data->ebooks = open_ebooks();
 
-	ret = e_book_async_get_contacts(ebook, query, cache_cb, data);
-	e_book_query_unref(query);
-	if (ret != FALSE) {
-		g_free(data);
-		if (err)
-			*err = -EFAULT;
-		return NULL;
+	/* Add 0.vcf */
+	if (e_book_get_self(&me, &eb, &gerr) == FALSE) {
+		g_error_free(gerr);
+		goto next;
+	}
+
+	evcard = E_VCARD(me);
+
+	cname = evcard_name_attribute_to_string(evcard);
+	if (!cname)
+		cname = g_strdup("");
+
+	attrib = e_vcard_get_attribute(evcard, EVC_UID);
+	uid = e_vcard_attribute_get_value(attrib);
+	if (!uid)
+		uid = g_strdup("");
+
+	attrib = e_vcard_get_attribute(evcard, EVC_TEL);
+	if (attrib)
+		tel =  e_vcard_attribute_get_value(attrib);
+	else
+		tel = g_strdup("");
+
+	data->entry_cb(uid, 0, cname, NULL, tel, data->user_data);
+
+	data->count++;
+
+	g_free(cname);
+	g_free(uid);
+	g_free(tel);
+	g_object_unref(eb);
+
+next:
+	ebook = data->ebooks;
+	while (ebook != NULL) {
+		if (e_book_is_opened(ebook->data) == TRUE) {
+			ret = e_book_get_contacts_async(ebook->data, query,
+								cache_cb, data);
+			if (ret == TRUE)
+				data->queued_calls++;
+		}
+
+		ebook = ebook->next;
 	}
 
 	if (err)
-		*err = 0;
+		*err = (data->queued_calls == 0 ? -ENOENT : 0);
 
 	return data;
 }
